from qtstrap import QColor


class colors:
    navy = '#001f3f'
    blue = '#0074D9'
    aqua = '#7FDBFF'
    teal = '#39CCCC'
    olive = '#3D9970'
    green = '#2ECC40'
    lime = '#01FF70'
    yellow = '#FFDC00'
    orange = '#FF851B'
    red = '#FF4136'
    maroon = '#85144b'
    fuchsia = '#F012BE'
    purple = '#B10DC9'
    black = '#111111'
    gray = '#AAAAAA'
    silver = '#DDDDDD'
    white = '#FFFFFF'


class qcolors:
    navy = QColor('#001f3f')
    blue = QColor('#0074D9')
    aqua = QColor('#7FDBFF')
    teal = QColor('#39CCCC')
    olive = QColor('#3D9970')
    green = QColor('#2ECC40')
    lime = QColor('#01FF70')
    yellow = QColor('#FFDC00')
    orange = QColor('#FF851B')
    red = QColor('#FF4136')
    maroon = QColor('#85144b')
    fuchsia = QColor('#F012BE')
    purple = QColor('#B10DC9')
    black = QColor('#111111')
    gray = QColor('#AAAAAA')
    silver = QColor('#DDDDDD')
    white = QColor('#FFFFFF')


from qtstrap import *


def run():
    # create the fundamental Qt objects
    app = BaseApplication()
    window = BaseMainWindow()

    # make everything a little easier to see
    set_font_options(window, {'setPointSize': 12})
    window.setMinimumSize(400, 300)

    # create the two widgets for our app
    label = QLabel('Hello World!', visible=False)
    button = QPushButton('Click me!')

    # create a function to use as an event handler
    def on_click():
        button.setVisible(False)
        label.setVisible(True)

    # register the handler to the event
    button.clicked.connect(on_click)

    # build a layout for our two widgets
    with CVBoxLayout(window) as layout:
        with layout.hbox(align='center') as layout:
            layout.add(label)
        layout.add(button)

    window.show() # show the window
    app.exec_() # start the Qt event loop


if __name__ == "__main__":
    run()


import time


session_start_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())


class Column:
    def __init__(self, name=None, title=None, visible=True, width=50):
        self.name = name
        self.title = title if title else name
        self.visible = visible
        self.width = width

    def set_visibility(self, visible):
        self.visible = visible


default_columns = [
    Column("TimeStamp", "Time", width=150, visible=True),
    Column("Source", width=200, visible=True),
    Column("LogLevel", "Level#", width=100, visible=False),
    Column("LogLevelName", "Level", width=80, visible=True),
    Column("Message", width=100, visible=True),
    Column("Args", width=100, visible=False),
    Column("Module", width=100, visible=False),
    Column("FuncName", "Func", width=100, visible=False),
    Column("LineNo", "Line", width=100, visible=False),
    Column("Exception", width=100, visible=False),
    Column("Process", width=100, visible=False),
    Column("Thread", width=100, visible=False),
    Column("ThreadName", width=100, visible=False),
]


class LogProfile:
    def __init__(self):
        self.columns = default_columns

        self.loggers = {
            'global': ['DEBUG','INFO','WARNING','ERROR','CRITICAL'],
        }
        self.visible_loggers = ['global']
        self.text_filter = ''
        self.current_session_only = True
        self.query_limit = 1000

    @property
    def visible_columns(self):
        return [c for c in self.columns if c.visible]

    def get_log_level_column(self):
        return [c.name for c in self.visible_columns].index('LogLevelName')

    def set_filter(self, filt):
        self.text_filter = filt['text']
        self.visible_loggers = filt['visible_loggers']
        self.loggers = filt['loggers']
        self.current_session_only = filt['current_session_only']
        self.query_limit = filt['query_limit']

    def build_query(self):
        query = "SELECT "

        columns = []
        for col in self.columns:
            if col.visible:
                if col.name != col.title:
                    columns.append(f'"{col.name}" AS "{col.title}"')
                else:
                    columns.append(f'"{col.name}"')

        query += ', '.join(columns)
        query += " FROM 'log'"

        where = []
        if self.current_session_only:
            where.append(f"TimeStamp > '{session_start_time}'")

        if self.text_filter:
            where.append(f"Message LIKE '%{self.text_filter}%'")
        
        if self.visible_loggers:
            sources = []
            for logger in self.visible_loggers:
                levels = ', '.join([f'"{l}"' for l in self.loggers[logger]])
                s = f'(Source = "{logger}" AND LogLevelName IN ({levels}))'
                sources.append(s)
            
            where.append(f"({' OR '.join(sources)})")

        if where:
            query += f' WHERE ' + ' AND '.join(where)
        
        if self.query_limit:
            query += f" LIMIT {self.query_limit}"

        return query


from qtstrap import *


class LinkLabel(QLabel):
    def __init__(self, text='', link='', both=None, parent=None):
        super().__init__(parent=parent)

        if both:
            self._text = both
            self._link = both
        else:
            self._text = text
            self._link = link

        self.setTextInteractionFlags(Qt.TextBrowserInteraction)
        self.setOpenExternalLinks(True)

        self._update_text()

    def setText(self, text):
        self._text = text
        self._update_text()

    def setLink(self, link):
        self._link = link
        self._update_text()

    def _update_text(self):
        super().setText(f'<a href="{self._text}">{self._link}</a>')


from qtstrap import *
from qtpy.QtSql import *
import time
from .log_filter_controls import log_levels
from .log_profile import LogProfile, Column, default_columns
from .log_database_handler import db_conn_name


session_start_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())


class LogDbModel(QSqlQueryModel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.log_level_column = 2

    def data(self, index, role=Qt.DisplayRole):
        if role == Qt.ForegroundRole:
            try:
                if index.column() == self.log_level_column:
                    level = super().data(index, Qt.DisplayRole)
                    return QColor(log_levels[level])
            except (IndexError, KeyError):
                pass
            return None
        else:
            return super().data(index, role)


class LogTableView(QTableView):
    def __init__(self):
        super().__init__()
        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.verticalHeader().hide()
        self.verticalHeader().setDefaultSectionSize(12)

        header = self.horizontalHeader()
        header.setStretchLastSection(True)
        header.setContextMenuPolicy(Qt.CustomContextMenu)
        header.customContextMenuRequested.connect(self.header_menu)
        header.setSectionsMovable(True)
        header.viewport().installEventFilter(self)

        self.db_model = LogDbModel(self)
        
        self.profile = LogProfile()

        self.setModel(self.db_model)
        self.need_to_refresh = False
        
        self.scan_timer = QTimer()
        self.scan_timer.timeout.connect(self.attempt_refresh)
        self.scan_timer.start(250)

    def attempt_refresh(self):
        if self.need_to_refresh:
            self.db_model.log_level_column = self.profile.get_log_level_column()
            self.refresh()
            self.need_to_refresh = False

    def schedule_refresh(self):
        self.need_to_refresh = True

    def set_filter(self, filt):
        self.profile.set_filter(filt)
        self.refresh()

    def refresh(self):
        at_bottom = False
        if self.verticalScrollBar().value() == self.verticalScrollBar().maximum():
            at_bottom = True
        
        self.db_model.setQuery(self.profile.build_query(), QSqlDatabase.database(db_conn_name))
        while self.db_model.canFetchMore():
            self.db_model.fetchMore()

        for i, column in enumerate(self.profile.visible_columns):
            self.horizontalHeader().resizeSection(i, column.width)

        if at_bottom:
            self.scrollToBottom()

    def eventFilter(self, object, event):
        if event.type() == QEvent.MouseButtonRelease:
            self.mouse_released()
            return True
        return False

    def mouse_released(self):
        for section in range(self.horizontalHeader().count()):
            col = self.profile.visible_columns[section]
            col.width = self.horizontalHeader().sectionSize(section)

    def header_menu(self):
        menu = QMenu()

        for column in self.profile.columns:
            action = QAction(column.title, menu, checkable=True)
            action.setChecked(column.visible)
            action.triggered.connect(column.set_visibility)
            menu.addAction(action)

        if menu.exec_(QCursor.pos()):
            self.need_to_refresh = True


from .options import *
from .settings import *
from .qt import *
from .utils import *
from .widgets import *
from .base_application import BaseApplication, install_app_info, install_ctrlc_handler
from .base_window import BaseMainWindow
from .toolbar import BaseToolbar, SettingsToolbar
from .timestamp import TimeStamp, time_since


# create a shorter alias for accessing the current QApplication
App = QApplication.instance


from qtstrap import *


class StateButton(QPushButton):
    state_changed = Signal(int)

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state):
        if self._state != state:
            self._state = state
            self.state_changed.emit(self._state)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setIconSize(QSize(60, 60))
        self.icons = []
        self._state = None
        
        self.clicked.connect(self.next_state)
        self.state_changed.connect(self.update_icon)

    def next_state(self):
        self.state = (self.state + 1) % len(self.icons)

    def update_icon(self):
        if self.icons:
            self.setIcon(self.icons[self.state])


class IconToggleButton(QPushButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setIconSize(QSize(60, 60))
        self.setCheckable(True)

        self.icon_unchecked = QIcon()
        self.icon_checked = QIcon()

        self.clicked.connect(self.update_icon)

    def update_icon(self):
        if self.isChecked():
            self.setIcon(self.icon_checked)
        else:
            self.setIcon(self.icon_unchecked)


class ConfirmToggleButton(QPushButton):
    _state_changed = Signal(int)

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state):
        if self._state != state:
            self._state = state
            self.timer.stop()
            self._state_changed.emit(self._state)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setIconSize(QSize(50, 50))
        self.icons = []

        self._state = None
        self.timer = QTimer(self)
        self.timer.timeout.connect(self.uncheck)
        self.timeout_time = 1000

        self.clicked.connect(self.next_state)
        self._state_changed.connect(self.update_icon)

    def uncheck(self):
        self.state = 0
        self.setChecked(False)
        self.setCheckable(False)

    def confirm(self):
        self.state = 1
        self.timer.start(self.timeout_time)

    def check(self):
        self.state = 2
        self.setCheckable(True)
        self.setChecked(True)

    def next_state(self):
        if self.state == 0:
            self.confirm()
        elif self.state == 1:
            self.check()
        elif self.state == 2:
            self.uncheck()

    def update_icon(self):
        if self.icons:
            self.setIcon(self.icons[self.state])


import sys
import click
from pathlib import Path
import shutil
from PyInquirer import prompt, print_json, Separator
import uuid
import os


@click.group()
def main():
    pass


def validate_name(answer):
    if len(answer) == 0:
        return "name can't be empty"
    if " " in answer:
        return "name can't contain spaces"
    return True


init_prompt = [
    {
        'type': 'input',
        'name': 'app_name',
        'message': 'Application Name:',
        'validate': validate_name,
    },
    {
        'type': 'input',
        'name': 'app_publisher',
        'message': 'Publisher Name:',
        'validate': validate_name,
    },
]


@main.command()
def init():
    parent = Path(__file__).parent 
    template = Path(parent / 'template')
    template_files = template.rglob('*')

    # copy template files to current working directory
    for f in template_files:
        dest = f.relative_to(template)
        if not dest.exists():
            if f.is_dir():
                Path.mkdir(dest)
            if f.is_file():
                shutil.copy(f, dest)

    # create app_info.py based on user input
    if not Path('src/app_info.py').exists():
        result = prompt(init_prompt)
        if result:
            with open('src/app_info.py', 'w') as f:
                f.write(f'AppName = "{result["app_name"]}"\n')
                f.write(f'AppVersion = "0.1"\n')
                f.write(f'AppPublisher = "{result["app_publisher"]}"\n')
                f.write(f'AppExeName = "{result["app_name"]}.exe"\n')
                f.write('AppIconName = "resources/application.ico"\n')
                guid = '{{' + str(uuid.uuid4()) + '}'
                f.write(f'AppId = "{guid}"\n')


@main.command()
def version():
    print('qtstrap v<something>')


# @main.command()
# def build():
#     print('build')





from .layouts import (
    ContextLayout,
    CVBoxLayout,
    CHBoxLayout,
    CGridLayout,
)
from .splitter import CSplitter, PersistentCSplitter
from .buttons import StateButton, IconToggleButton, ConfirmToggleButton
from .labeledit import LabelEdit
from .link_label import LinkLabel
from .line_widgets import VLine, HLine
from .persistent_widgets import (
    PersistentCheckBox,
    PersistentLineEdit,
    PersistentTextEdit,
    PersistentListWidget,
    PersistentTreeWidget,
    PersistentComboBox,
    PersistentCheckableAction,
)
from .persistent_tab_widget import PersistentTabWidget


from qtstrap import *


class Widget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.button = QPushButton('button')
        self.label = QLabel('label')

        with CVBoxLayout(self) as layout:
            layout.add(self.button)
            layout.add(self.label)


class MainWindow(BaseMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.button = QPushButton('button')
        self.label = QLabel('label')
        self.widget = Widget()

        with CVBoxLayout(self) as layout:
            layout.add(self.button)
            layout.add(self.label)
            layout.add(self.widget)


def test_enable_and_disable_children(qtbot):
    window = MainWindow()
    
    # children should start enabled
    assert window.button.isEnabled() == True
    assert window.label.isEnabled() == True
    assert window.widget.button.isEnabled() == True
    assert window.widget.label.isEnabled() == True

    disable_children(window)
    assert window.button.isEnabled() == False
    assert window.label.isEnabled() == False
    assert window.widget.button.isEnabled() == False
    assert window.widget.label.isEnabled() == False

    enable_children(window)
    assert window.button.isEnabled() == True
    assert window.label.isEnabled() == True
    assert window.widget.button.isEnabled() == True
    assert window.widget.label.isEnabled() == True

    disable_children(window.widget)
    assert window.button.isEnabled() == True
    assert window.label.isEnabled() == True
    assert window.widget.button.isEnabled() == False
    assert window.widget.label.isEnabled() == False

    enable_children(window.widget)
    assert window.button.isEnabled() == True
    assert window.label.isEnabled() == True
    assert window.widget.button.isEnabled() == True
    assert window.widget.label.isEnabled() == True


def test_get_all_children(qtbot):
    widget = Widget()
    children = get_children(widget)

    # this Widget should have 3 children: a layout, a pushbutton, and a label
    assert len(children) == 3
    assert type(children[0]) == CVBoxLayout
    assert type(children[1]) == QPushButton
    assert type(children[2]) == QLabel


def test_set_font_options(qtbot):
    window = MainWindow()
    
    # default font size is 8
    assert window.label.font().pointSize() == 8
    assert window.label.font().bold() == False

    set_font_options(window, {'setPointSize': 12, 'setBold': True})

    assert window.label.font().pointSize() == 12
    assert window.label.font().bold() == True


from qtstrap import *


def test_context_layout():
    widget = QWidget()

    with CVBoxLayout(widget) as layout:
        layout.add(QLabel('test'))

    assert len(get_children(widget)) == 2


def test_nested_context_layout():
    widget = QWidget()

    with CVBoxLayout(widget) as layout:
        with layout.hbox() as layout:
            layout.add(QLabel('upper left'))
            layout.add(QLabel('upper right'))
        with layout.hbox() as layout:
            layout.add(QLabel('lower left'))
            layout.add(QLabel('lower right'))

    assert len(get_children(widget)) == 7


from qtstrap import *
import pytest


class MainWindow(BaseMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.button = QPushButton('Click me!', clicked=lambda: self.label.setText('clicked'))
        self.label = QLabel('^^^^^^^')

        with CVBoxLayout(self) as layout:
            layout.add(self.button)
            layout.add(self.label)


class Application(BaseApplication):
    def __init__(self) -> None:
        super().__init__()

        self.window = MainWindow()


@pytest.fixture(scope="session")
def qapp():
    yield Application()


def test_base_application_startup(qtbot, qapp):
    qtbot.addWidget(qapp.window)
    assert qapp.window.label.text() == '^^^^^^^'

    qtbot.mouseClick(qapp.window.button, QtCore.Qt.LeftButton)
    assert qapp.window.label.text() == 'clicked'


from qtstrap import *
from .splitter import CSplitter, PersistentCSplitter


alignments = {
    'left': Qt.AlignLeft,
    'right': Qt.AlignRight,
    'top': Qt.AlignTop,
    'bot': Qt.AlignBottom,
    'bottom': Qt.AlignBottom,
    'center': Qt.AlignCenter,
}


class ContextLayout:
    def __init__(self, parent=None, stretch=None, margins=None, align=None, **kwargs):
        if isinstance(parent, QMainWindow):
            # a layout can't be added directly to a QMainWindow, because it requires a "centralWidget"
            # we can use a throwaway QWidget as a shim
            widget = QWidget(parent=parent)
            parent.setCentralWidget(widget)
            super().__init__(widget, **kwargs)
        elif isinstance(parent, QSplitter):
            # a layout can't be added directly to a QSplitter, only a widget
            # we can use a throwaway QWidget as a shim
            widget = QWidget(parent=parent)
            parent.addWidget(widget)
            super().__init__(widget, **kwargs)
        elif parent is None or isinstance(parent, QWidget):
            super().__init__(parent, **kwargs)
        else:
            super().__init__(**kwargs)
            if stretch:
                parent.addLayout(self, stretch)
            else:
                parent.addLayout(self)

        if margins:
            self.setContentsMargins(*margins)

        if align in alignments:
            self.setAlignment(alignments[align])

        self._stack = []
        self.next_layout = None

    def __getattr__(self, name):
        return getattr(self._layout, name)

    @property
    def _layout(self):
        layout = self
        if len(self._stack) > 0:
            layout = self._stack[len(self._stack) - 1]
        return layout

    def __call__(self):
        return self._layout

    def add(self, item, *args, **kwargs):
        if isinstance(item, QWidget):
            self._layout.addWidget(item, *args, **kwargs)
        elif isinstance(item, QLayout):
            self._layout.addLayout(item, *args, **kwargs)
        elif isinstance(item, list):
            for i in item:
                self._layout.add(i, *args, **kwargs)

        return item

    def vbox(self, *args, **kwargs):
        self.next_layout = CVBoxLayout(self._layout, *args, **kwargs)
        return self

    def hbox(self, *args, **kwargs):
        self.next_layout = CHBoxLayout(self._layout, *args, **kwargs)
        return self

    def grid(self, *args, **kwargs):
        self.next_layout = CGridLayout(self._layout, *args, **kwargs)
        return self

    def split(self, name=None, **kwargs):
        if name:
            self.next_layout = PersistentCSplitter(name, self._layout, **kwargs)
        else:
            self.next_layout = CSplitter(self._layout, **kwargs)
        return self

    def __enter__(self):
        if self.next_layout is not None:
            self.next_layout.__enter__()
            self._stack.append(self.next_layout)
            self.next_layout = None
        return self

    def __exit__(self, *_):
        if len(self._stack) > 0:
            item = self._stack.pop()
            item.__exit__()


class CVBoxLayout(ContextLayout, QVBoxLayout):
    pass


class CHBoxLayout(ContextLayout, QHBoxLayout):
    pass


class CGridLayout(ContextLayout, QGridLayout):
    pass


from qtstrap import *


class LabelEdit(QWidget):
    text_changed = Signal(str)

    def __init__(self, text, *args, changed=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.setToolTip('doubleclick to edit')

        self.label = QLabel(text)
        self.edit = QLineEdit()
        self.edit.installEventFilter(self)

        self.stack = QStackedLayout(self)
        self.stack.insertWidget(0, self.label)
        self.stack.insertWidget(1, self.edit)

        if changed:
            self.text_changed.connect(changed)

    def text(self):
        return self.label.text()

    def setText(self, text):
        self.label.setText(text)
        self.text_changed.emit(self.edit.text())

    def start_editing(self):
        self.edit.setText(self.label.text())
        self.edit.setFocus()
        self.stack.setCurrentIndex(1)

    def mouseDoubleClickEvent(self, event: PySide2.QtGui.QMouseEvent) -> None:
        self.start_editing()
        return super().mouseDoubleClickEvent(event)

    def accept(self):
        self.setText(self.edit.text())
        self.stack.setCurrentIndex(0)

    def dismiss(self):
        self.stack.setCurrentIndex(0)

    def eventFilter(self, source, event):
        if event.type() == QEvent.KeyPress:
            if event.key() == Qt.Key_Return:
                self.accept()
                return True
            
            if event.key() == QtCore.Qt.Key_Escape:
                self.dismiss()
                event.accept()
                return True

        return False


from .qt import *
from pathlib import Path


class BaseMainWindow(QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
            
        self.setObjectName("MainWindow")
        self.load_settings()

    def closeEvent(self, event):
        self.save_settings()
        super().closeEvent(event)
        
    def save_settings(self):
        QSettings().setValue("mainwindow/geometry", self.saveGeometry())
        QSettings().setValue("mainwindow/state", self.saveState())

    def load_settings(self):
        if QSettings().value("mainwindow/geometry") is not None:
            self.restoreGeometry(QSettings().value("mainwindow/geometry"))
        if QSettings().value("mainwindow/state") is not None:
            self.restoreState(QSettings().value("mainwindow/state"))


import sys


def uncache(exclude):
    """Remove package modules from cache except excluded ones.
    On next import they will be reloaded.
    
    Args:
        exclude (iter<str>): Sequence of module paths.
    """
    pkgs = []
    for mod in exclude:
        pkg = mod.split('.', 1)[0]
        pkgs.append(pkg)

    to_uncache = []
    for mod in sys.modules:
        if mod in exclude:
            continue

        if mod in pkgs:
            to_uncache.append(mod)
            continue

        for pkg in pkgs:
            if mod.startswith(pkg + '.'):
                to_uncache.append(mod)
                break

    for mod in to_uncache:
        del sys.modules[mod]


from PySide2 import QtCore


class PortableSettings(QtCore.QSettings):
    settings_file_path = 'settings.ini'

    def __init__(self, *args, **kwargs):
        super().__init__(self.settings_file_path, QtCore.QSettings.IniFormat, *args, **kwargs)

    @staticmethod
    def _install():
        QtCore.QSettings = PortableSettings
        uncache(['QtCore'])


from qtstrap.options import OPTIONS


if OPTIONS.portable:
    PortableSettings.settings_file_path = OPTIONS.PORTABLE_SETTINGS_FILE.as_posix()
    PortableSettings._install()


from qtstrap import *
import qtawesome as qta
from qtstrap.extras.style import colors
import json


log_levels = {
    'DEBUG': colors.aqua,
    'INFO': colors.green,
    'WARNING': colors.yellow,
    'ERROR': colors.orange,
    'CRITICAL': colors.red,
}
level_map = {
    'D': 'DEBUG',
    'I': 'INFO',
    'W': 'WARNING',
    'E': 'ERROR',
    'C': 'CRITICAL',
}


class LoggerDelegate(QStyledItemDelegate):
    def paint(self, painter, option, index):
        self.initStyleOption(option, index)
        value = index.data(Qt.DisplayRole)
        checked = index.data(Qt.UserRole)

        if value is None:
            return

        painter.save()

        if len(value) == 1:
            if option.state & QStyle.State_Selected and checked:
                painter.setPen(QPen(log_levels[level_map[value]]))
            else: 
                painter.setPen(QPen('gray'))
            painter.drawText(option.rect, Qt.AlignCenter, value)
        else:
            if option.state & QStyle.State_Selected:
                painter.setPen(QPen('lightgray'))
            else: 
                painter.setPen(QPen('gray'))
            painter.drawText(option.rect, Qt.AlignLeft, value)

        painter.restore()


class LoggerTreeWidgetItem(QTreeWidgetItem):
    def __init__(self, parent, name, full_name):
        super().__init__(parent)

        self.setText(0, name)
        self.name = name
        self.full_name = full_name

        self.levels = {
            'DEBUG': True,
            'INFO': True,
            'WARNING': True,
            'ERROR': True,
            'CRITICAL': True,
        }

        self.update_data()
        self.selected = False

    def clicked(self, column):
        if column == 0:
            if self.full_name != 'global':
                self.setSelected(not self.isSelected())
            else:
                self.selected = not self.selected
        else:
            if self.data(column, Qt.UserRole):
                self.setData(column, Qt.UserRole, False)
                self.levels[level_map[self.text(column)]] = False
            else:
                self.setData(column, Qt.UserRole, True)
                self.levels[level_map[self.text(column)]] = True

    def double_clicked(self, column):        
        def select_children(item, state):
            item.setSelected(state)
            for i in range(item.childCount()):
                select_children(item.child(i), state)

        if column == 0:
            if self.full_name != 'global':
                state = self.isSelected()
            else:
                state = self.selected
            
            for i in range(self.childCount()):
                select_children(self.child(i), state)

    def update_data(self):
        for i, level in enumerate(self.levels):
            self.setData(i + 1, Qt.UserRole, self.levels[level])
            self.setText(i + 1, level[:1])
            self.setTextAlignment(i + 1, Qt.AlignCenter)

    def set_levels(self, level_filter=[]):
        for level in level_filter:
            if level in self.levels:
                self.levels[level] = True

        self.update_data()

    def get_levels(self):
        return [level for level in self.levels if self.levels[level]]

    def set_all_levels(self, state: bool):
        for level in self.levels:
            self.levels[level] = state
        self.update_data()


class LoggerTreeWidget(QTreeWidget):
    filter_updated = Signal()

    def __init__(self):
        super().__init__()
        
        self.setItemDelegate(LoggerDelegate())
        self.setSelectionMode(QAbstractItemView.NoSelection)
        self.setRootIsDecorated(False)
        self.setIndentation(10)
        self.setStyleSheet("QTreeView::branch { border-image: url(none.png); }" );
        self.setUniformRowHeights(True)
        self.setExpandsOnDoubleClick(False)
        self.setItemsExpandable(False)
        self.setFocusPolicy(Qt.NoFocus)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.header().setMinimumSectionSize(1)
        self.header().hide()

        self.setColumnCount(7)
        self.setColumnWidth(1,15)
        self.setColumnWidth(2,15)
        self.setColumnWidth(3,15)
        self.setColumnWidth(4,15)
        self.setColumnWidth(5,15)

        self.loggers = {}
        self.visible_loggers = []

        self.itemClicked.connect(self.click)
        self.itemDoubleClicked.connect(self.double_click)
        self.itemSelectionChanged.connect(self.selection_changed)

        self.root = LoggerTreeWidgetItem(self, 'global', 'global')
        self.loggers['global'] = self.root
        self.root.setSelected(True)

    def click(self, item, column):
        item.clicked(column)
        self.selection_changed()

    def double_click(self, item, column):
        item.double_clicked(column)
        self.selection_changed()

    def selection_changed(self):
        self.visible_loggers = [item.full_name for item in self.selectedItems()]
        self.filter_updated.emit()

    def register_logger(self, full_name):
        if full_name in self.loggers:
            return 
        else:
            parts = full_name.rsplit('.', 1)  # split off the last name only
            name = parts[-1]

            if len(parts) == 1:
                parent = self.root
            else:
                if parts[0] not in self.loggers:
                    self.register_logger(parts[0])
                parent = self.loggers[parts[0]]

            self.loggers[full_name] = LoggerTreeWidgetItem(parent, name, full_name)
            self.loggers[full_name].setSelected(True)

        self.expandAll()
        self.selection_changed()

    def register_loggers(self, loggers):
        for name in loggers:
            self.register_logger(name)

    def set_visible_loggers(self, logger_filter):
        for logger in logger_filter:
            self.setItemSelected(self.loggers[logger], True)

    def contextMenuEvent(self, event):
        menu = QMenu()
        pos = event.globalPos()
        menu.addAction(QAction('Select Only', menu, triggered=lambda: self.select_only(pos)))
        menu.addAction(QAction('Select All', menu, triggered=self.select_all))
        menu.addAction(QAction('Deselect All', menu, triggered=self.deselect_all))
        menu.addAction(QAction('All Levels', menu, triggered=lambda: self.enable_all_levels(pos)))
        menu.addAction(QAction('No Levels', menu, triggered=lambda: self.disable_all_levels(pos)))
        menu.addAction(QAction('Enable Everything', menu, triggered=self.enable_everything))
        menu.addAction(QAction('Disable Everything', menu, triggered=self.disable_everything))
        menu.exec_(event.globalPos())

    def set_levels_of_children(self, item, state):
        if hasattr(item, 'set_all_levels'):
            item.set_all_levels(state)
        for i in range(item.childCount()):
            self.set_levels_of_children(item.child(i), state)

    def enable_everything(self):
        self.select_all()
        self.set_levels_of_children(self.invisibleRootItem(), True)

    def disable_everything(self):
        self.deselect_all()
        self.set_levels_of_children(self.invisibleRootItem(), False)

    def enable_all_levels(self, pos):
        self.itemAt(self.viewport().mapFromGlobal(pos)).set_all_levels(True)
        self.selection_changed()

    def disable_all_levels(self, pos):
        self.itemAt(self.viewport().mapFromGlobal(pos)).set_all_levels(False)
        self.selection_changed()

    def set_visible_loggers(self, visible_loggers):
        def set_visible(item):
            if hasattr(item, 'full_name'):
                if item.full_name in visible_loggers:
                    item.setSelected(True)
                else:
                    item.setSelected(False)

            for i in range(item.childCount()):
                set_visible(item.child(i))

        set_visible(self.invisibleRootItem())

    def select_all(self):
        def select_children(item):
            item.setSelected(True)
            for i in range(item.childCount()):
                select_children(item.child(i))

        select_children(self.invisibleRootItem())

    def deselect_all(self):
        self.clearSelection()
        self.root.setSelected(True)

    def select_only(self, pos):
        self.deselect_all()
        self.itemAt(self.viewport().mapFromGlobal(pos)).setSelected(True)
        self.selection_changed()


class ProfileSelector(QWidget):
    added = Signal(str)
    removed = Signal(str)
    changed = Signal(str)

    def __init__(self):
        super().__init__()
        self.setStyleSheet('QPushButton { max-width: 20px; }')

        self.selector = QComboBox()
        self.editor = QLineEdit()

        self.add = QPushButton(qta.icon('fa.plus-square-o', color='lightgray'), '')
        self.accept = QPushButton(qta.icon('fa5.check-square', color='lightgray'), '')
        self.edit = QPushButton(qta.icon('fa.pencil-square-o', color='lightgray'), '')

        self.selector.currentIndexChanged.connect(self.on_change)
        self.add.clicked.connect(self.on_add)
        self.accept.clicked.connect(self.on_accept)
        self.editor.returnPressed.connect(self.on_accept)
        
        grid = CGridLayout(self, margins=(0, 0, 0, 0), spacing=2)

        grid.add(self.selector, 0, 0, 1, 3)
        grid.add(self.editor, 0, 0, 1, 3)
        grid.add(self.add, 0, 3)
        grid.add(self.accept, 0, 3)
        grid.add(self.edit, 0, 4)

        self.accept.hide()
        self.editor.hide()

    def on_change(self):
        name = self.selector.currentText()
        self.changed.emit(name)

    def on_add(self):
        self.accept.show()
        self.add.hide()
        self.selector.hide()
        self.editor.show()
        self.editor.setFocus()

    def on_accept(self):
        self.accept.hide()
        self.add.show()
        self.selector.show()
        self.editor.hide()

        name = self.editor.text()
        if len(name) > 0:
            self.added.emit(name)
        self.editor.clear()

    def on_remove(self):
        name = self.selector.currentText()
        self.removed.emit(name)


class FilterControls(QStackedWidget):
    filter_updated = Signal(dict)

    empty_profile = {
        'loggers': {
            'global': ['DEBUG','INFO','WARNING','ERROR','CRITICAL'],
        },
        'visible_loggers': [
            'global'
        ],
        'text': '',
        'current_session_only': True,
        'query_limit': 1000,
    }

    default_settings = {
        'selected_profile': 'default',
        'registered_loggers': ['global'],
        'profiles': {
            'default': empty_profile
        }
    }

    def __init__(self):
        super().__init__()

        self.setStyleSheet("""
            QTreeWidget {
                selection-background-color: transparent;
                selection-color: lightgray; 
                color: gray;
            }
        """)
        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)

        # create widgets
        self.profiles = ProfileSelector()

        self.text_filter = QLineEdit()
        self.text_filter.textChanged.connect(self.update_filter)
        self.text_filter.setClearButtonEnabled(True)
        self.text_filter.setPlaceholderText('filter by text')

        self.logger_filter = LoggerTreeWidget()
        self.session_checkbox = QCheckBox()
        self.query_limit = QLineEdit()

        # load settings and send filter components to widgets
        prev = QSettings().value('log_monitor/log_settings', json.dumps(self.default_settings))
        self.settings = json.loads(prev)

        profiles = self.settings['profiles']
        current_profile_name = self.settings['selected_profile']
        if current_profile_name not in profiles:
            current_profile_name = list(profiles.keys())[0]
        self.current_profile = profiles[current_profile_name]

        self.logger_filter.register_loggers(self.settings['registered_loggers'])
        self.profiles.selector.addItems(profiles)

        for logger in self.current_profile['loggers']:
            self.logger_filter.register_logger(logger)

        self.logger_filter.set_visible_loggers(self.current_profile['visible_loggers'])
        self.session_checkbox.setChecked(self.current_profile['current_session_only'])
        self.query_limit.setText(str(self.current_profile['query_limit']))

        # connect signals
        self.logger_filter.filter_updated.connect(self.update_filter)
        self.profiles.selector.setCurrentIndex(self.profiles.selector.findText(current_profile_name))
        self.profiles.changed.connect(self.change_profile)
        self.profiles.added.connect(self.add_profile)
        self.profiles.removed.connect(self.remove_profile)
        self.profiles.edit.clicked.connect(lambda: self.setCurrentIndex(1))
        self.session_checkbox.stateChanged.connect(self.update_filter)

        # send the filter to the model
        self.update_filter()

        self.addWidget(QWidget())
        self.addWidget(QWidget())

        # controls layout
        with CVBoxLayout(self.widget(0), margins=(0, 0, 0, 0)) as layout:
            layout.add(self.profiles)
            with layout.hbox() as layout:
                layout.add(QLabel('Current Session:'))
                layout.add(self.session_checkbox)
                layout.add(QLabel(), 1)
            with layout.hbox() as layout:
                layout.add(QLabel('Query Limit:'))
                layout.add(self.query_limit)
                layout.add(QLabel(), 1)
            layout.add(self.text_filter)
            layout.add(self.logger_filter)

        # editor layout
        with CVBoxLayout(self.widget(1), margins=(0, 0, 0, 0)) as layout:
            with layout.hbox() as layout:
                layout.add(QLabel(), 1)
                layout.add(QPushButton('X', maximumWidth=20, clicked=lambda: self.setCurrentIndex(0)))
            layout.add(QLabel(), 1)

    def save_settings(self):
        QSettings().setValue('log_monitor/log_settings', json.dumps(self.settings))

    def add_profile(self, name):
        new_profile = dict(self.empty_profile)
        self.settings['profiles'][name] = new_profile
        self.profiles.selector.addItem(name)
        self.profiles.selector.setCurrentIndex(self.profiles.selector.findText(name))

    def remove_profile(self, name):
        index = self.profiles.selector.findText(name)
        self.profiles.selector.removeItem(index)

        self.settings['profiles'].pop(name)
        self.save_settings()

    def change_profile(self, profile_name):
        self.settings['selected_profile'] = profile_name

        self.current_profile = self.settings['profiles'][profile_name]
        self.logger_filter.set_visible_loggers(self.current_profile['visible_loggers'])
        self.update_filter()
        self.save_settings()

    def update_filter(self):
        text = self.text_filter.text()
        loggers = {item.full_name: item.get_levels() for _, item in self.logger_filter.loggers.items()}
        visible_loggers = self.logger_filter.visible_loggers

        self.settings['registered_loggers'] = list(self.logger_filter.loggers.keys())
        self.current_profile['text'] = text
        self.current_profile['loggers'] = loggers
        self.current_profile['visible_loggers'] = visible_loggers
        self.current_profile['current_session_only'] = self.session_checkbox.isChecked()
        self.current_profile['query_limit'] = 1000

        self.filter_updated.emit(self.current_profile)
        self.save_settings()



from qtstrap import *


darkPalette = QStyleFactory.create('fusion').standardPalette()


light = QColor('#525252')
mid = QColor('#414141')
dark = QColor('#313131')
accent= QColor('#ca3e47')


# base
darkPalette.setColor(QPalette.WindowText, QColor(180, 180, 180))
darkPalette.setColor(QPalette.Button, QColor(53, 53, 53))
darkPalette.setColor(QPalette.Light, QColor(180, 180, 180))
darkPalette.setColor(QPalette.Midlight, QColor(90, 90, 90))
darkPalette.setColor(QPalette.Dark, QColor(35, 35, 35))
darkPalette.setColor(QPalette.Text, QColor(180, 180, 180))
darkPalette.setColor(QPalette.BrightText, QColor(180, 180, 180))
darkPalette.setColor(QPalette.ButtonText, QColor(180, 180, 180))
darkPalette.setColor(QPalette.Base, QColor(42, 42, 42))
darkPalette.setColor(QPalette.Window, QColor(53, 53, 53))
darkPalette.setColor(QPalette.Shadow, QColor(20, 20, 20))
darkPalette.setColor(QPalette.Highlight, QColor(42, 130, 218))
darkPalette.setColor(QPalette.HighlightedText, QColor(180, 180, 180))
darkPalette.setColor(QPalette.Link, QColor(56, 252, 196))
darkPalette.setColor(QPalette.AlternateBase, QColor(66, 66, 66))
darkPalette.setColor(QPalette.ToolTipBase, QColor(53, 53, 53))
darkPalette.setColor(QPalette.ToolTipText, QColor(180, 180, 180))

# disabled
darkPalette.setColor(QPalette.Disabled, QPalette.WindowText, QColor(127, 127, 127))
darkPalette.setColor(QPalette.Disabled, QPalette.Text, QColor(127, 127, 127))
darkPalette.setColor(QPalette.Disabled, QPalette.ButtonText, QColor(127, 127, 127))
darkPalette.setColor(QPalette.Disabled, QPalette.Highlight, QColor(80, 80, 80))
darkPalette.setColor(QPalette.Disabled, QPalette.HighlightedText, QColor(127, 127, 127))



from qtstrap import *


class VLine(QFrame):
    def __init__(self, parent=None, line_width=1):
        super().__init__(parent=parent)
        self.setFrameShape(QFrame.VLine)
        # expand vertically, but not horizontally
        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
        self.setLineWidth(line_width)


class HLine(QFrame):
    def __init__(self, parent=None, line_width=1):
        super().__init__(parent=parent)
        self.setFrameShape(QFrame.HLine)

        # expand horizontally, but not vertically
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.setLineWidth(line_width)


from qt import *
import typing
from style import colors
import re


class CommandRegistry(QObject):
    def __init__(self) -> None:
        self.registry = {}
        self.commands = []

    def register_command(self, command):
        self.registry[command.text()] = command
        self.commands.append(command)
        self.commands.sort(key=lambda x: x.text())

    def execute(self, command_name):
        self.registry[command_name].triggered.emit()


registry = CommandRegistry()


class Command(QAction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        registry.register_command(self)

        self.usage_count = 0
        self.triggered.connect(self.used)
    
    def used(self):
        self.usage_count += 1


class PopupDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.prefix = ""
        self.normal = QPen('gray')
        self.contains = QPen('lightgray')
        self.highlight = QPen(QColor(colors.blue))

    def set_prefix(self, prefix):
        self.prefix = prefix

    def paint(self, 
            painter: PySide2.QtGui.QPainter, 
            option: PySide2.QtWidgets.QStyleOptionViewItem, 
            index: PySide2.QtCore.QModelIndex
        ):

        self.initStyleOption(option, index)
        prefix = self.prefix
        value = index.data(Qt.EditRole)
        shortcut = index.data(Qt.UserRole)

        # adjust full drawing area
        option.rect.setX(option.rect.x() + 5)
        option.rect.setWidth(option.rect.width() - 10)

        painter.save()

        if option.state & QStyle.State_Selected:
            painter.fillRect(option.rect, QColor('#243F89'))

        if prefix == "":
            painter.setPen(self.normal)
            painter.drawText(option.rect, Qt.AlignLeft, value)
        else:
            if prefix.lower() in value.lower():
                parts = re.split(prefix, value, flags=re.IGNORECASE)
                
                # the split is case insensitive, so use the lengths of the
                # parts to slice the original text out of the complete string
                sections = [parts[0]]
                length = len(parts[0])
                for part in parts[1:]:
                    sections.append(value[length:length + len(prefix)])
                    sections.append(part)
                    length += len(prefix) + len(part)

                prev = None
                rect = QRect(option.rect)
                for text in sections:
                    if text.lower() == prefix.lower():
                        painter.setPen(self.highlight)
                    else:
                        painter.setPen(self.contains)

                    if prev:
                        rect = QRect(prev.x() + prev.width(), prev.y(), option.rect.width(), prev.height())

                    prev = painter.drawText(rect, Qt.AlignLeft, text)
            else:
                painter.setPen(self.normal)
                painter.drawText(option.rect, Qt.AlignLeft, value)

        painter.setPen(self.normal)
        painter.drawText(option.rect, Qt.AlignRight, shortcut)

        painter.restore()


class CommandModel(QAbstractListModel):
    sorted_commands = []

    def sort_commands(self, prefix):
        self.sorted_commands = [cmd for cmd in registry.commands if prefix.lower() in cmd.text().lower()]
        result = bool(self.sorted_commands)
        self.sorted_commands.extend([cmd for cmd in registry.commands if prefix.lower() not in cmd.text().lower()])
        return result

    def rowCount(self, parent: PySide2.QtCore.QModelIndex) -> int:
        return len(registry.commands)

    def data(self, index: PySide2.QtCore.QModelIndex, role: int) -> typing.Any:
        if not index.isValid():
            return None

        if role == Qt.EditRole:
            return self.sorted_commands[index.row()].text()
        
        elif role == Qt.UserRole:
            return self.sorted_commands[index.row()].shortcut().toString()

    def index(self, row: int, column: int, parent: PySide2.QtCore.QModelIndex) -> PySide2.QtCore.QModelIndex:
        return self.createIndex(row, column)


class CommandCompleter(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.list = QListView()
        self.list.setUniformItemSizes(True)
        self.list.setSelectionRectVisible(True)
        self.list.setSelectionMode(QAbstractItemView.SingleSelection)
        self.list.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.list.setFixedHeight(500)

        self.command_model = CommandModel(self)
        self.list.setModel(self.command_model)

        self.delegate = PopupDelegate(self)
        self.list.setItemDelegate(self.delegate)
        
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.list)
        self.active = False

    def open(self):
        self.active = True
        self.update_prefix('')
        super().show()

    def close(self):
        self.active = False
        self.update_prefix('')
        super().hide()

    def update_prefix(self, prefix):
        self.delegate.set_prefix(prefix)
        self.command_model.sort_commands(prefix)

        index = self.list.model().index(0, 0, QModelIndex())
        self.list.setCurrentIndex(index)
        
        # redraw items in popup
        for row in range(self.list.model().rowCount(QModelIndex())):
            index = self.list.model().index(row, 0, QModelIndex())
            self.list.update(index)

    def move_selection_up(self):
        current = self.list.currentIndex()

        if current.row() > 0:
            new = self.list.model().index(current.row() - 1, 0, QModelIndex())
            self.list.setCurrentIndex(new)

    def move_selection_down(self):
        current = self.list.currentIndex()

        if current.row() < self.list.model().rowCount(QModelIndex()) - 1:
            new = self.list.model().index(current.row() + 1, 0, QModelIndex())
            self.list.setCurrentIndex(new)

    def get_selection(self):
        index = self.list.currentIndex()
        return index.data(Qt.EditRole)


class _CommandPalette(QDialog):
    _instance = None

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setObjectName('CommandPalette')
        self.setWindowFlags(Qt.Tool | Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)
        self.setFocusPolicy(Qt.StrongFocus)
        
        set_font_options(self, {'setPointSize': 16})
        
        self.setMinimumWidth(700)

        self.action = QAction("Command Palette", self)
        self.action.setShortcut('Ctrl+Shift+P')
        self.action.triggered.connect(self.palette)

        self.line = QLineEdit()
        self.line.setStyleSheet("""
            QLineEdit {
                font-size: 16pt;
                border: 1px solid #0074D9;
            }
        """)
        self.command_completer = CommandCompleter(self)

        self.line.textChanged.connect(self.command_completer.update_prefix)
        self.line.returnPressed.connect(self.accept)
        self.command_completer.list.clicked.connect(self.accept)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(5, 5, 5, 5)
        layout.addWidget(self.line)
        layout.addWidget(self.command_completer)

        self.command_completer.close()

        self.installEventFilter(self)
        self.line.installEventFilter(self)
        self.callback = None

    def palette(self):
        self.open()
        self.command_completer.open()

    def _open(self, cb=None, prompt=None, placeholder=None, completer=None, validator=None, mask=None):
        self.callback = cb

        self.line.setText(prompt)
        self.line.setPlaceholderText(placeholder)
        self.line.setCompleter(completer)
        self.line.setValidator(validator)
        self.line.setInputMask(mask)

        self.center_on_parent()
        self.show()
        self.activateWindow()
        self.line.setFocus()

    def open(self, cb=None, prompt=None, placeholder=None, completer=None, validator=None, mask=None):
        QTimer.singleShot(0, lambda: self._open(cb, prompt, placeholder, completer, validator, mask))

    def accept(self):
        if self.command_completer.active:
            name = self.command_completer.get_selection()
            self.dismiss()
            registry.execute(name)
        else:
            result = self.line.text()
            if self.callback:
                self.callback(result)
            self.dismiss()

    def dismiss(self):
        self.callback = None
        self.line.clear()
        self.line.setPlaceholderText('')
        self.line.setCompleter(None)
        self.line.setValidator(None)
        self.line.setInputMask('')
        self.hide()
        self.command_completer.close()

    def eventFilter(self, source, event):
        if event.type() == QEvent.KeyPress:
            if event.key() == Qt.Key_Return:
                self.accept()
                
            if source is not self:
                if self.command_completer.active:
                    if event.key() == Qt.Key_Up:
                        event.accept()
                        self.command_completer.move_selection_up()
                        return True

                    elif event.key() == Qt.Key_Down:
                        event.accept()
                        self.command_completer.move_selection_down()
                        return True
            
            if event.key() == QtCore.Qt.Key_Escape:
                self.dismiss()
                event.accept()
                return True

        if event.type() == QEvent.WindowDeactivate:
            self.dismiss()
            event.accept()
            return True

        return False

    def center_on_parent(self):
        r = self.parent().frameGeometry()
        rect = QRect(r.x() - (self.width() / 2), r.y(), r.width(), 100)
        self.move(rect.center())


def CommandPalette(parent=None):
    if _CommandPalette._instance is None:
        _CommandPalette._instance = _CommandPalette(parent)

    return _CommandPalette._instance


from time import time


def time_since(timestamp):
    return (time() - timestamp)


class TimeStamp:
    def __init__(self):
        self._time = time()

    def time_since(self, timestamp):
        return (time() - timestamp)

    def update(self):
        self._time = time()


from qtstrap import *


class PersistentCheckBox(QCheckBox):
    def __init__(self, name, changed=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name
        self.restore_state()

        if changed:
            self.stateChanged.connect(changed)

        self.stateChanged.connect(lambda: QSettings().setValue(self.name, self.checkState()))
    
    def restore_state(self):
        prev_state = QSettings().value(self.name, 0)
        if prev_state == int(Qt.Checked):
            self.setCheckState(Qt.Checked)
        elif prev_state == int(Qt.PartiallyChecked):
            self.setCheckState(Qt.PartiallyChecked)


class PersistentLineEdit(QLineEdit):
    def __init__(self, name, *args, default='', changed=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name
        self.default = default
        self.restore_state()

        if changed:
            self.textChanged.connect(changed)

        self.textChanged.connect(lambda: QSettings().setValue(self.name, self.text()))
    
    def restore_state(self):
        self.setText(str(QSettings().value(self.name, self.default)))


class PersistentTextEdit(QTextEdit):
    def __init__(self, name, *args, default='', changed=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name
        self.default = default
        self.restore_state()

        if changed:
            self.textChanged.connect(changed)

        self.textChanged.connect(lambda: QSettings().setValue(self.name, self.toPlainText()))
    
    def restore_state(self):
        self.setText(str(QSettings().value(self.name, self.default)))


class PersistentListWidget(QListWidget):
    def __init__(self, name, items=[], default=[], changed=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name
        self.default_selection = default

        if items:
            self.addItems(items)
            self.restore_state()
                
        if changed:
            self.itemSelectionChanged.connect(changed)

        self.itemSelectionChanged.connect(lambda: QSettings().setValue(self.name, self.selected_items()))

    def selected_items(self):
        return [item.text() for item in self.selectedItems()]

    def restore_state(self):
        prev_items = QSettings().value(self.name, self.default_selection)
        if prev_items:
            for i in range(self.count()):
                if self.item(i).text() in prev_items:
                    self.item(i).setSelected(True)


class PersistentTreeWidget(QTreeWidget):
    def __init__(self, name, items=[], index_column=0, default=[], changed=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name
        self.default_selection = default
        self.index_column = index_column

        if items:
            self.addItems(items)
            self.restore_state()
                
        if changed:
            self.itemSelectionChanged.connect(changed)

        self.itemSelectionChanged.connect(lambda: QSettings().setValue(self.name, self.selected_items()))

    def selected_items(self):
        return [item.text(self.index_column) for item in self.selectedItems()]

    def restore_state(self):
        prev_items = QSettings().value(self.name, self.default_selection)
        if prev_items:
            for i in range(self.count()):
                if self.item(i).text(self.index_column) in prev_items:
                    self.item(i).setSelected(True)


class PersistentComboBox(QComboBox):
    def __init__(self, name, items=[], changed=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name

        if items:
            self.addItems(items)
            self.restore_state()

        if changed:
            self.currentTextChanged.connect(changed)

        self.currentTextChanged.connect(lambda: QSettings().setValue(self.name, self.currentIndex()))

    def restore_state(self):
        prev_index = QSettings().value(self.name, 0)
        if isinstance(prev_index, int):
            self.setCurrentIndex(prev_index)


class PersistentCheckableAction(QAction):
    def __init__(self, name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name
        self.setCheckable(True)
        self.restore_state()

        self.triggered.connect(lambda: QSettings().setValue(self.name, self.isChecked()))

    def restore_state(self):
        prev_state = QSettings().value(self.name, 0)
        if prev_state == 'true':
            self.setChecked(True)
        elif prev_state == 'false':
            self.setChecked(False)


from .log_widget import LogMonitorWidget, LogMonitorDockWidget, LogMonitorDropdown
from .log_database_handler import DatabaseHandler
import logging
from qtstrap import OPTIONS
from pathlib import Path


# Make sure the log database directory exists
Path(OPTIONS.config_dir).mkdir(parents=True, exist_ok=True)

db_path = OPTIONS.config_dir + '/log.db'


def install(database_name=db_path):
    logger = logging.getLogger()
    logger.setLevel(1)
    logger.addHandler(DatabaseHandler(database_name))


from qtstrap import *


class MainWindow(BaseMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.button = QPushButton('Click me!', clicked=lambda: self.label.setText('clicked'))
        self.label = QLabel('^^^^^^^')

        with CVBoxLayout(self) as layout:
            layout.add(self.button)
            layout.add(self.label)


def test_base_window_startup(qtbot):
    window = MainWindow()

    qtbot.addWidget(window)
    assert window.label.text() == '^^^^^^^'

    qtbot.mouseClick(window.button, QtCore.Qt.LeftButton)
    assert window.label.text() == 'clicked'


# def test_base_window_settings(qtbot):
#     window = MainWindow()

#     qtbot.addWidget(window)
#     assert window.geometry() == QRect(0, 0, 640, 480)

#     window.setGeometry(100, 100, 640, 480)
#     assert window.geometry() == QRect(100, 100, 640, 480)

#     window.close()
    
#     window = MainWindow()
    
#     qtbot.addWidget(window)
#     assert window.geometry() == QRect(100, 100, 640, 480)


from qtstrap import OPTIONS
from .qt import *
import signal
from pathlib import Path


def install_ctrlc_handler(app):
    def ctrlc_handler(sig=None, frame=None):
        app.closeAllWindows() # this makes sure the MainWindow's .close() method gets called
        app.quit()
       
    # grab the keyboard interrupt signal 
    signal.signal(signal.SIGINT, ctrlc_handler)

    # empty timer callback
    def update():
        pass
    
    # create timer to force python interpreter to get some runtime
    app._ctrlc_timer = QTimer()
    app._ctrlc_timer.timeout.connect(update)
    app._ctrlc_timer.start(10)


def install_app_info(app):
    if OPTIONS.app_info:
        info = OPTIONS.app_info

        if info.AppPublisher:
            app.setOrganizationName(info.AppPublisher)
        if info.AppPublisher:
            app.setOrganizationDomain(info.AppPublisher)
        if info.AppName:
            app.setApplicationName(info.AppName)
        if info.AppVersion:
            app.setApplicationVersion(info.AppVersion)

        if info.AppIconName and Path(info.AppIconName).exists():
                app.setWindowIcon(QIcon(info.AppIconName))


class BaseApplication(QApplication):
    def __init__(self, *args, register_ctrlc_handler=True, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        install_app_info(self)

        if register_ctrlc_handler:
            install_ctrlc_handler(self)


from qtstrap import *
from qtpy.QtSql import *
from .log_table_view import LogTableView
from .log_filter_controls import FilterControls
from .log_database_handler import DatabaseHandler, db_conn_name


try:
    from command_palette import CommandPalette, Command
    command_palette_available = True
except:
    command_palette_available = False


class LogMonitorWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.tab_name = "Log Monitor"

        if command_palette_available:
            self.commands = [
                Command("Log Monitor: Switch profile", triggered=self.open_profile_prompt),
            ]

        self.log_table = LogTableView()
        DatabaseHandler.register_callback(self.log_table.schedule_refresh)

        self.filter_controls = FilterControls()
        self.filter_controls.filter_updated.connect(self.log_table.set_filter)
        self.filter_controls.update_filter()

        self.query_existing_loggers()

        with CPersistentSplitter('log_monitor_splitter', self) as splitter:
            splitter.add(self.filter_controls, 1)
            splitter.add(self.log_table, 10)

    def open_profile_prompt(self):
        profiles = list(self.filter_controls.settings['profiles'].keys())

        self.completer = QCompleter(self, profiles)
        self.completer.setCaseSensitivity(Qt.CaseInsensitive)
        self.completer.setCompletionMode(QCompleter.PopupCompletion)

        QTimer.singleShot(0, lambda: CommandPalette().open(
            placeholder='Select a profile',
            cb=lambda result: print(result),
            completer=self.completer
        ))

    def query_existing_loggers(self):
        db = QSqlDatabase.database(db_conn_name)
        query = db.exec_("SELECT Source FROM 'log'")
        loggers = set()
        while query.next():
            loggers.add(query.value(0))
        self.filter_controls.logger_filter.register_loggers(loggers)


class LogMonitorDockWidget(QDockWidget):
    def __init__(self, parent=None):
        super().__init__('Log Monitor', parent=parent)
        self.setObjectName('LogMonitor')

        self.setWidget(LogMonitorWidget(self))

        if command_palette_available:
            self.commands = [
                Command("Log Monitor: Show log monitor", triggered=self.show, shortcut='Ctrl+L'),
                Command("Log Monitor: Hide log monitor", triggered=self.hide),
            ]

        self.setAllowedAreas(Qt.AllDockWidgetAreas)
        self.dockLocationChanged.connect(lambda: QTimer.singleShot(0, self.adjust_size))

        self.starting_area = Qt.BottomDockWidgetArea

        if not self.parent().restoreDockWidget(self):
            self.parent().addDockWidget(self.starting_area, self)
        
        self.closeEvent = lambda x: self.hide()

    def adjust_size(self):
        if self.isFloating():
            self.adjustSize()
            
    def toggleViewAction(self):
        action = super().toggleViewAction()
        action.setShortcut('Ctrl+L')
        return action


class LogMonitorDropdown(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setWindowFlags(Qt.Tool | Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)
        self.setFocusPolicy(Qt.StrongFocus)

        CHBoxLayout(self).add(LogMonitorWidget(self))
        
        if command_palette_available:
            self.commands = [
                Command("Log Monitor: Show log monitor", triggered=self.show, shortcut='Ctrl+L'),
                Command("Log Monitor: Hide log monitor", triggered=self.hide),
            ]
        
        self.hide()

    def toggleViewAction(self):
        action = QAction("Toggle Log Monitor", self)
        action.setShortcut('`')
        action.triggered.connect(self.toggle_view)
        return action

    def toggle_view(self):
        if self.isVisible():
            self.hide()
        else:
            self.center_on_parent()
            self.show()

    def center_on_parent(self):
        offset = 33
        r = self.parent().frameGeometry()
        rect = QRect(r.x(), r.y() + offset, r.width(), r.height() - (offset * 2))
        self.setGeometry(rect)


from .qt import *
import qtawesome as qta


location_map = {
    'left': Qt.LeftToolBarArea,
    'right': Qt.RightToolBarArea,
    'top': Qt.TopToolBarArea,
    'bottom': Qt.BottomToolBarArea,
}


class BaseToolbar(QToolBar):
    def __init__(self, parent=None, name=None, location=None, size=40):
        super().__init__(parent=parent)

        self.setMovable(False)
        self.setIconSize(QSize(size, size))
        if name:
            self.setObjectName(name)
        if location and parent:
            if isinstance(location, str):
                location = location_map[location]
            parent.addToolBar(location, self)

    def add_spacer(self):
        empty = QWidget(self)
        empty.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding);
        self.addWidget(empty)


class SettingsToolbar(BaseToolbar):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.settings_button = QToolButton(self, icon=qta.icon('fa.gear', color='gray'))
        self.addWidget(self.settings_button)

        self.settings_menu = QMenu(self.settings_button)
        self.settings_button.setMenu(self.settings_menu)
        self.settings_button.setPopupMode(QToolButton.InstantPopup)

        self.quit_action = QAction(
            '&Exit', self.settings_menu,
            shortcut='Ctrl+Q',
            statusTip='Exit application',
            triggered=self.close
        )

    def add_action(self, action):
        self.settings_menu.addAction(action)

    def add_separator(self):
        self.settings_menu.addSeparator()



from qtstrap import *


class PersistentTabWidget(QTabWidget):
    def __init__(self, name, tabs=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name
        
        self.movable = False
        if 'movable' in kwargs:
            self.movable = kwargs['movable']
            self.tabBar().tabMoved.connect(self.move_tab)

        if tabs:
            if isinstance(tabs, list):
                tab_dict = {}
                for tab in tabs:
                    if hasattr(tab, 'tab_name'):
                        tab_dict[tab.tab_name] = tab
                tabs = tab_dict

            self.restore_state(tabs)

        self.setTabsClosable(True)
        self.tabCloseRequested.connect(self.close_tab)

        self.tabBar().setContextMenuPolicy(Qt.CustomContextMenu)
        self.tabBar().customContextMenuRequested.connect(self.menu)

    def menu(self, pos) -> None:
        menu = QMenu()
        menu.addAction(QAction('Unhide All', self, triggered=self.unhide_all))
        menu.exec_(self.tabBar().mapToGlobal(pos))

    def unhide_all(self):
        for i in range(self.count()):
            self.setTabVisible(i, True)
        QSettings().setValue(self.name + '/visibility', self.visible_tabs())

    def restore_state(self, tabs=None):
        visible_tabs = QSettings().value(self.name + '/visibility', {})

        def add_tab(tab, name):
            self.addTab(tab, name)
            if name in visible_tabs:
                if not visible_tabs[name]:
                    self.setTabVisible(self.count() - 1, False)

        if tabs:
            if self.movable:
                added_tabs = {}
                if prev_tabs := QSettings().value(self.name + '/order', []):
                    for tab in prev_tabs:
                        added_tabs[tab] = tabs[tab]
                        add_tab(tabs[tab], tab)
                    for tab in tabs:
                        if tab not in prev_tabs:
                            add_tab(tabs[tab], tab)
                else:
                    for name, tab in tabs.items():
                        add_tab(tab, name)
            else:
                for name, tab in tabs.items():
                    add_tab(tab, name)

        prev_index = QSettings().value(self.name + '/current', 0)
        if isinstance(prev_index, int):
            self.setCurrentIndex(min(int(prev_index), self.count()))

        self.currentChanged.connect(lambda i: QSettings().setValue(self.name + '/current', i))

    def tab_order(self):
        order = []
        for i in range(self.count()):
            order.append(self.tabText(i))
        return order

    def visible_tabs(self):
        visible = {}
        for i in range(self.count()):
            visible[self.tabText(i)] = self.isTabVisible(i)
        return visible

    def move_tab(self, old, new):
        QSettings().setValue(self.name + '/order', self.tab_order())

    def close_tab(self, index):
        self.setTabVisible(index, False)
        QSettings().setValue(self.name + '/visibility', self.visible_tabs())



from qtstrap import *
from .highlighters import PythonHighlighter


class CodeEditor(QTextEdit):
    def __init__(self, *args, highlighter=PythonHighlighter, **kwargs):
        super().__init__(*args, **kwargs)
        
        set_font_options(self, {
            'setFamily': 'Courier',
            'setStyleHint': QFont.Monospace,
            'setFixedPitch': True,
        })

        # font = QFont();
        # font.setFamily("Courier");
        # font.setStyleHint(QFont.Monospace);
        # font.setFixedPitch(True);
        # self.setFont(font)
        
        self.setTabStopWidth(QFontMetricsF(self.font()).width(' ') * 4)
        self.syntax = highlighter(self)


import logging
from qtstrap import *
from qtpy.QtSql import *
import time


db_conn_name = 'logs'


initial_sql = """
CREATE TABLE IF NOT EXISTS log(
    TimeStamp TEXT,
    Source TEXT,
    LogLevel INT,
    LogLevelName TEXT,
    Message TEXT,
    Args TEXT,
    Module TEXT,
    FuncName TEXT,
    LineNo INT,
    Exception TEXT,
    Process INT,
    Thread TEXT,
    ThreadName TEXT
)
"""

insertion_sql = """
INSERT INTO log(
    TimeStamp,
    Source,
    LogLevel,
    LogLevelName,
    Message,
    Args,
    Module,
    FuncName,
    LineNo,
    Exception,
    Process,
    Thread,
    ThreadName
)
VALUES (
    '%(dbtime)s',
    '%(name)s',
    %(levelno)d,
    '%(levelname)s',
    '%(msg)s',
    '%(args)s',
    '%(module)s',
    '%(funcName)s',
    %(lineno)d,
    '%(exc_text)s',
    %(process)d,
    '%(thread)s',
    '%(threadName)s'
);
"""


class DatabaseHandler(logging.Handler):
    callbacks = []

    """A logging.Handler subclass that redirects outbound records to a local sqlite3 database """
    def __init__(self, database_name):
        super().__init__()
        self.formatter = logging.Formatter("%(asctime)s")
        
        db = QSqlDatabase.addDatabase('QSQLITE', db_conn_name)
        db.setDatabaseName(database_name)
        db.open()
        db.exec_(initial_sql)

    def format_time(self, record):
        record.dbtime = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(record.created))

    def emit(self, record):
        self.format(record)
        self.format_time(record)
        
        if record.exc_info:  # for exceptions
            record.exc_text = logging._defaultFormatter.formatException(record.exc_info)
        else:
            record.exc_text = ""

        # Insert the log record
        try:
            QSqlDatabase.database(db_conn_name).exec_(insertion_sql % record.__dict__)
        except ValueError:
            pass

        for cb in self.callbacks:
            cb()

    @classmethod
    def register_callback(cls, cb):
        cls.callbacks.append(cb)

    def write(self, m):
        pass


from qtstrap import *
from .layouts import *


class CSplitter(QSplitter):
    def __init__(self, parent=None, orientation=None, **kwargs):
        if isinstance(parent, QWidget):
            super().__init__(parent, **kwargs)
            CHBoxLayout(parent).add(self)
        elif isinstance(parent, QLayout):
            super().__init__(**kwargs)
            parent.add(self)
        
        if orientation:
            self.setOrientation(orientation)

    def add(self, item, stretch=None):
        if isinstance(item, QWidget):
            self.addWidget(item)
            if stretch:
                self.setStretchFactor(self.count()-1, stretch)
        elif isinstance(item, QLayout):
            self.addWidget(QWidget(self, layout=item))
            if stretch:
                self.setStretchFactor(self.count()-1, stretch)
        elif isinstance(item, list):
            for i in item:
                self.add(i)
        return item

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


class PersistentCSplitter(CSplitter):
    def __init__(self, name, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self.name = name

        self.splitterMoved.connect(lambda: QSettings().setValue(self.name, self.saveState()))
    
    def restore_state(self):
        if state := QSettings().value(self.name, None):
            self.restoreState(state)

    def __exit__(self, *args):
        self.restore_state()


from qtstrap import *


def enable_children(thing: QObject) -> None:
    """
    Recursively walk the provided thing and enable all of its widget children.
    """
    # QMainWindows don't have children, they have a centralWidget
    if isinstance(thing, QMainWindow):
        thing = thing.centralWidget()

    # QWidgets don't have children, they have a layout that has children
    if isinstance(thing, QWidget):
        thing = thing.layout()

    for i in range(thing.count()):
        if thing.itemAt(i).widget():
            thing.itemAt(i).widget().setEnabled(True)
        else:
            enable_children(thing.itemAt(i))


def disable_children(thing: QObject) -> None:
    """
    Recursively walk the provided thing and disable all of its widget children.
    """
    # QMainWindows don't have children, they have a centralWidget
    if isinstance(thing, QMainWindow):
        thing = thing.centralWidget()

    # QWidgets don't have children, they have a layout that has children
    if isinstance(thing, QWidget):
        thing = thing.layout()

    for i in range(thing.count()):
        if thing.itemAt(i).widget():
            thing.itemAt(i).widget().setEnabled(False)
        else:
            disable_children(thing.itemAt(i))


def get_children(obj: QObject) -> list:
    """
    Recursively visit all the children of the specified object and collect them in a list.
    """
    children = []

    def _get_children(obj, prefix=''):
        for child in obj.children():
            children.append(child)
            _get_children(child, '  ' + prefix )

    _get_children(obj)

    return children


def print_children(obj: QObject, prefix='') -> None:
    """
    Recursively visit all the children of the specified object and print them.
    """
    for child in obj.children():
        print(prefix, child)
        print_children(child, '  ' + prefix )


def set_font_options(obj: QObject, options={}):
    """
    Set the QFont options of the specified object.
    Font options are specified by providing the name of the QFont setter method.
    
    Example:
    set_font_options(widget, {'setPointSize': 12, 'setBold': True})

    is equivalent to writing:
    font = widget.font()
    font.setPointSize(12)
    font.setBold(True)
    widget.setFont(font)
    """
    font = obj.font()
    for setting, value in options.items():
        getattr(font, setting)(value)
    obj.setFont(font)
    return obj


from pathlib import Path
import sys
from appdirs import AppDirs


class OPTIONS:
    APPLICATION_PATH = Path(sys.argv[0]).parent.absolute()
    app_info_available = False
    app_info = None
    portable = False
    PORTABLE_SETTINGS_FILE = Path(APPLICATION_PATH / 'settings.ini')
    PORTABLE_FLAG_FILE = Path(APPLICATION_PATH / '.portable')


# attempt to import app_info.py
try:
    import app_info
    OPTIONS.app_info_available = True
    OPTIONS.app_info = app_info
except ModuleNotFoundError:
    OPTIONS.app_info_available = False


# check for the file ".portable"
if Path(OPTIONS.PORTABLE_FLAG_FILE).exists():
    OPTIONS.portable = True


# set up appdirs
if not OPTIONS.portable and OPTIONS.app_info:
    dirs = AppDirs(app_info.AppName, app_info.AppPublisher)
    OPTIONS.config_dir = dirs.user_config_dir
else:
    OPTIONS.config_dir = OPTIONS.APPLICATION_PATH.as_posix()